IMPLEMENTATION MODULE Dlg;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **************************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    --------------------------------------------------------------------------
    D.Maslov    Ver  1.0  Initial version                          February 88
    A.Iakovlev  Ver  1.1  Use new RasterDisplay                    March 88
    A.Iakovlev  Ver  1.2  Revised keys processing                  August 88
    A.Iakovlev  Ver  1.3  Use new SimpleDrawings                   October 88
    A.Iakovlev  Ver  1.4  Use new RasterDisplay                    November 88
    D.Maslov    Ver  2.0  Full redesign                            June 89
    A.Iakovlev  Ver  2.1  Implementation of redesign               July 89
    A.Iakovlev  Ver  3.0  Menu attributes                          July 89
    A.Iakovlev  Ver  3.1  Use M2VERS utility                       August 89
    A.Iakovlev  Ver  3.2  Use new RasterDisplay                    September 89
    A.Iakovlev  Ver  4.0  Renaming                                 January 90
    A.Iakovlev  Ver  4.1  Use new module SrvLex                    April 90
    A.Iakovlev  Ver  4.2  Window-based menus                       June 90
    A.Iakovlev  Ver  4.3  New module Win                           August 90
    A.Iakovlev  Ver  4.4  Initialization of opaque type            August 90
    A.Iakovlev  Ver  4.5  New module ScrTxOut                      October 90
    A.Iakovlev  Ver  5.0  Full redesign                            April 91
    A.Iakovlev  Ver  5.1  Cosmetics in New                         June 92
    A.Iakovlev  Ver  6.0  Use module Res                           August 92
    A.Iakovlev  Ver  7.0  Pragma data(var_enum_size)               August 92
    A.Iakovlev  Ver  8.0  Renaming                                 June 93
    A.Iakovlev  Ver  8.1  Import from SYSTEM                       January 94
    A.Iakovlev  Ver  8.2  Use module SrvLow                        January 94
    A.Iakovlev  Ver  8.3  Use type SrvLex.CharSet                  January 94
    A.Iakovlev  Ver  8.4  Use SrvSys.Add instead of Lib.AddAddr    January 94
    A.Iakovlev  Ver  8.5  Use module Strings                       January 94
    A.Iakovlev  Ver  8.6  Use ScrTxOut.CorrectAttrs                February 94
    A.Iakovlev  Ver  8.7  Stony Brook support                      February 94
    A.Iakovlev  Ver  8.8  No use Lib.ScanR, Lib.ScanNeL            February 94
    A.Iakovlev  Ver  8.9  Temporary form on stack (was on heap)    March 94
    A.Iakovlev  Ver  8.10 Use module Mem for heap control          October 94
    A.Iakovlev  Ver  8.11 MaxNmbOfBoxes = 256                      March 95
    A.Iakovlev  Ver  9.0  No graphics - only TEXT box store !!!    March 95
    A.Iakovlev  Ver 10.0  New box type = runLine                   March 95
    A.Iakovlev  Ver 11.0  Simplifications in attributes            March 95
    **************************************************************************
    TopSpeed compiler dependencies:
    ------ pragmas (*#...*)
    ------ structured constants
    Stony Brook compiler dependencies:
    ------ structured constants
    TopSpeed library dependencies:
    ------ import from Strings: Length, Pos, Concat, CompareStr
    Stony Brook library dependencies:
    ------ import from Strings: Length, Pos, Concat, CompareStr
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT
    SYSTEM, Strings,
    DrvMem, Geo, Res, SrvSys, SrvLow, SrvLex, SrvObj,
    Scr, ScrOut, Win, WinOut;

  CONST
    (* ----------------------- Parameters of plexes / boxes *)
    MaxNmbOfBoxes = 256;
    MaxNmbOfParts = 48;
    MaxNmbOfItems = 12;
    MaxBoxStoreLen = 2000;
    MaxItemStrLen = 24;
    MarkStrLen = 4;
    (* ----------------------- Separators *)
    LS = "{";
    RS = "}";
    PC = "_";
    SP = " ";

  TYPE
    (* ----------------------------------------- Text store for box content *)
    BoxStoreString = ARRAY [0 .. MaxBoxStoreLen - 1] OF CHAR;
    BoxStorePtr = POINTER TO BoxStoreString;
    (* ------------------------------------------------ Parts of box store *)
    PartDescriptor =
      RECORD
        row: CARDINAL;              (* --- row number in text coordinates *)
        col: CARDINAL;              (* --- column number in text coordinates *)
        inx: CARDINAL;              (* --- beginning index in box store *)
        len: CARDINAL;              (* --- length of part *)
        pic: CARDINAL;              (* --- length of picture sub-part *)
        txt: BOOLEAN;               (* --- TRUE, iff non-item part *)
      END;
    PartArray = ARRAY [0 .. MaxNmbOfParts - 1] OF PartDescriptor;
    (* -------------------------------------------------------------- Items *)
    ItemString = ARRAY [0 .. MaxItemStrLen - 1] OF CHAR;
    ItemDescriptor =
      RECORD
        key: CARDINAL;              (* --- item key *)
        partNo: CARDINAL;           (* --- corresponding part number *)
        type: ItemType;             (* --- item type *)
        mark: BOOLEAN;              (* --- item mark *)
      END;
    ItemArray = ARRAY [0 .. MaxNmbOfItems - 1] OF ItemDescriptor;
    (* -------------------------------------------------------------- Boxes *)
    BoxDescriptor =
      RECORD
        win: Win.WINDOW;            (* --- box window *)
        basePt: Geo.Point;          (* --- base point for the window *)
        textWdt: CARDINAL;          (* --- window width in characters *)
        textHgt: CARDINAL;          (* --- window height in characters *)
        wdt: CARDINAL;              (* --- window width in pixels *)
        hgt: CARDINAL;              (* --- window height in pixels *)
        storePtr: BoxStorePtr;      (* --- pointer to box store *)
        storeLen: CARDINAL;         (* --- length of box store *)
        nmbOfParts: CARDINAL;       (* --- number of parts in box store *)
        part: PartArray;            (* --- parts array *)
        nmbOfItems: CARDINAL;       (* --- number of items in box store *)
        itemCurrent: CARDINAL;      (* --- current item number *)
        item: ItemArray;            (* --- items array *)
        type: BoxType;              (* --- box type *)
        position: BoxPosition;      (* --- box position *)
        winExist: BOOLEAN;          (* --- TRUE, iff box window exists *)
        runIndex: CARDINAL;         (* --- index of current character *)
        runDstPt: Geo.Point;        (* --- base point of current character *)
        runRect: Geo.Rectangle;     (* --- rolled rectangle *)
      END;
    BoxPtr = POINTER TO BoxDescriptor;
    BoxArray = ARRAY [0 .. MaxNmbOfBoxes - 1] OF BoxPtr;
    (* ---------------------------------------------------- Plex descriptor *)
    PLEX = POINTER TO PlexDescriptor;
    PlexDescriptor =
      RECORD
        next: PLEX;                 (* --- next plex *)
        attrs: Attrs;               (* --- attributes of the plex *)
        charWdt: CARDINAL;          (* --- width of character; in pixels *)
        charHgt: CARDINAL;          (* --- height of character; in pixels *)
        textAttrs: ScrOut.Attrs;    (* --- attributes of normal text *)
        itemAttrs: ScrOut.Attrs;    (* --- attributes of item text *)
        boxPtr: BoxArray;           (* --- box array *)
      END;
    (* ------------------------------------------------- *)
    MarkString = ARRAY [0 .. MarkStrLen - 1] OF CHAR;
    BoxTypeSet = SET OF BoxType;

  CONST
    BoxTypeStrLen = 10;
    BoxPosStrLen = 10;
    BoxStoreStrLen = 8;
    ItemTypeStrLen = 8;

  TYPE
    BoxTypeString = ARRAY [0 .. BoxTypeStrLen - 1] OF CHAR;
    BoxTypeArray = ARRAY BoxType OF BoxTypeString;
    (* ------------------------------------------------------- *)
    BoxPosString = ARRAY [0 .. BoxPosStrLen - 1] OF CHAR;
    BoxPosArray = ARRAY BoxPosition OF BoxPosString;
    (* ------------------------------------------------------- *)
    ItemTypeString = ARRAY [0 .. ItemTypeStrLen - 1] OF CHAR;
    ItemTypeArray = ARRAY ItemType OF ItemTypeString;
    (* ----------------------------------------------- *)
    DoubleChar = ARRAY [0 .. 1] OF CHAR;

  (*%T CMPTS*)

  CONST
    BT = BoxTypeArray  ("VERTICAL", "HORIZONTAL", "LINE", "FORM", "RUNLINE");
    BP = BoxPosArray   ("CENTERED", "TOP", "BOTTOM", "MOVABLE");
    IT = ItemTypeArray ("VIEW", "CALL", "CALLEXIT", "LINK");
    LB = DoubleChar    (15C, 12C);

  (*%E*)

  (*%T CMPSB*)

  CONST
    BT:BoxTypeArray  = ["VERTICAL", "HORIZONTAL", "LINE", "FORM", "RUNLINE"];
    BP:BoxPosArray   = ["CENTERED", "TOP", "BOTTOM", "MOVABLE"];
    IT:ItemTypeArray = ["VIEW", "CALL", "CALLEXIT", "LINK"];
    LB:DoubleChar    = [15C, 12C];

  (*%E*)

  VAR topPlex: PLEX;

  (* ================================================= Utilities *)

  PROCEDURE min(a, b: CARDINAL): CARDINAL;
    BEGIN
      IF a <= b THEN
        RETURN a;
      ELSE
        RETURN b;
      END;
    END min;

  PROCEDURE Caps(VAR s: ARRAY OF CHAR);
  (* --- translate all the characters of the strings into capitals *)
    VAR
      j: CARDINAL;
    BEGIN
      FOR j := 0 TO HIGH(s) DO s[j] := CAP(s[j]) END;
    END Caps;

  PROCEDURE FixPart(VAR p: PartDescriptor;
                    row, col, inx, len, pic: CARDINAL; txt: BOOLEAN);
  (* --- set the part description *)
    BEGIN
      p.row := row;
      p.col := col;
      p.inx := inx;
      p.len := len;
      p.pic := pic;
      p.txt := txt;
    END FixPart;

  (* ============================================ Plex utilities *)

  PROCEDURE NewPlex(a: Attrs; VAR pl: PLEX): BOOLEAN;
  (* --- create new empty plex and initialize its main parameters *)
    VAR
      newPlex: PLEX;
      j: CARDINAL;
    BEGIN
      IF DrvMem.Alloc(newPlex, SIZE(PlexDescriptor)) THEN
        WITH newPlex^ DO
          attrs := a;
          ScrOut.Dims(a.font, charWdt, charHgt);
          ScrOut.SetTextAttrs
          (
            textAttrs,
            Scr.copy, a.textColor, a.backColor,
            ScrOut.fillOpaqueStippled, a.font
          );
          ScrOut.SetTextAttrs
          (
            itemAttrs,
            Scr.copy, a.itemColor, a.backColor,
            ScrOut.fillOpaqueStippled, a.font
          );
          FOR j := 0 TO MaxNmbOfBoxes - 1 DO boxPtr[j] := NIL END;
        END;
        pl := newPlex;
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END NewPlex;

  PROCEDURE DisposeAllBoxes(pl: PLEX);
  (* --- dispose all the boxes in the plex *)
    VAR
      j: CARDINAL;
      res: Res.Status;
    BEGIN
      WITH pl^ DO
        FOR j := 0 TO MaxNmbOfBoxes - 1 DO
          IF boxPtr[j] # NIL THEN
            WITH boxPtr[j]^ DO
              IF storeLen > 0 THEN DrvMem.Free(storePtr, storeLen) END;
              IF winExist THEN res := Win.Dispose(win) END;
            END;
            DrvMem.Free(boxPtr[j], SIZE(BoxDescriptor));
          END;
        END;
      END;
    END DisposeAllBoxes;

  (* =================================================== Box utilities *)

  PROCEDURE NewBox(pl: PLEX; n: CARDINAL; t: BoxType; p: BoxPosition): BOOLEAN;
  (* --- create new empty box and initialize some its parameters *)
    VAR
      b: BoxPtr;
    BEGIN
      IF DrvMem.Alloc(b, SIZE(BoxDescriptor)) THEN
        WITH b^ DO
          storeLen := 0;
          nmbOfParts := 0;
          nmbOfItems := 0;
          itemCurrent := 0;
          type := t;
          position := p;
          winExist := FALSE;
        END;
        pl^.boxPtr[n] := b;
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END NewBox;

  PROCEDURE NewBoxStore(pl: PLEX; n: CARDINAL; sp: BoxStorePtr; m: CARDINAL): BOOLEAN;
  (* --- create box store in the box and copy the memory block into it *)
    BEGIN
      IF m > MaxBoxStoreLen THEN m := MaxBoxStoreLen END;
      WITH pl^.boxPtr[n]^ DO
        IF DrvMem.Alloc(storePtr, m) THEN
          SrvLow.MoveInc(sp, storePtr, m);
          storeLen := m;
          RETURN TRUE;
        ELSE
          RETURN FALSE;
        END;
      END;
    END NewBoxStore;

  PROCEDURE NewBoxWindow(pl: PLEX; n: CARDINAL): BOOLEAN;
  (* --- create the graphics box window and clear it with background color *)
    VAR
      a: Win.Attrs;
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          (* ---------------------- set window attributes *)
          a.basePt := basePt;
          a.wdt := wdt;
          a.hgt := hgt;
          a.frameAttrs := attrs.frameAttrs;
          a.titleAttrs := attrs.titleAttrs;
          (* --------------------------- attempt to create the window *)
          IF Win.New(a, win) = Res.done THEN
            (* -------------------------------------- success: clear it *)
            WinOut.Clear(win, attrs.backColor);
            (* --------------------------------------- normal return *)
            RETURN TRUE;
          ELSE
            (* ------------- fail during the window creation *)
            RETURN FALSE;
          END;
        END;
      END;
    END NewBoxWindow;

  PROCEDURE SetBoxExtent(pl: PLEX; n: CARDINAL);
    VAR
      scrWdt, scrHgt, frmWdt, frmHgt: CARDINAL;
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          (* -------------------- get screen and frame dimensions *)
          Scr.Dims(scrWdt, scrHgt);
          Win.FrameDims(attrs.frameAttrs, frmWdt, frmHgt);
          (* -------------------------------- compute the window extent *)
          IF type = line THEN
            wdt := scrWdt - 2 * frmWdt;
            hgt := attrs.topGap + charHgt + attrs.bottomGap;
          ELSE
            wdt := attrs.leftGap + textWdt * charWdt + attrs.rightGap;
            hgt := attrs.topGap + textHgt * charHgt + attrs.bottomGap;
          END;
          (* --------------------------------------- compute the base point *)
          basePt.ix := (VAL(INTEGER, scrWdt) - VAL(INTEGER, wdt)) DIV 2;
          CASE position OF
          | top:
            basePt.iy := VAL(INTEGER, frmHgt) + VAL(INTEGER, Win.TitleHgt(attrs.titleAttrs));
          | bottom:
            basePt.iy := VAL(INTEGER, scrHgt) - VAL(INTEGER, frmHgt) - VAL(INTEGER, hgt);
          ELSE
            basePt.iy := (VAL(INTEGER, scrHgt) - VAL(INTEGER, hgt)) DIV 2;
          END;
        END;
      END;
    END SetBoxExtent;

  (* ================================================== Item utilities *)

  PROCEDURE AddItem(pl: PLEX; n, p, k: CARDINAL; t: ItemType): BOOLEAN;
  (* --- add the item to the box description *)
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems < MaxNmbOfItems THEN
          WITH item[nmbOfItems] DO
            key := k;
            partNo := p;
            type := t;
            mark := FALSE;
          END;
          INC(nmbOfItems);
          RETURN TRUE;
        ELSE
          RETURN FALSE;
        END;
      END;
    END AddItem;

  PROCEDURE FindItem(pl: PLEX; n: CARDINAL; s: ItemString; VAR p: CARDINAL): BOOLEAN;
  (* --- find the item in the box store and return its part number *)
    VAR
      itemInx, partNo: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        itemInx := Strings.Pos(s, storePtr^);
        IF itemInx < MAX(CARDINAL) THEN
          partNo := 0;
          WHILE partNo < nmbOfParts DO
            IF part[partNo].inx = itemInx THEN
              p := partNo;
              RETURN TRUE;
            ELSE
              INC(partNo);
            END;
          END;
        END;
        RETURN FALSE;
      END;
    END FindItem;

  (* ================================================ Form utilities *)

  PROCEDURE PrepareBoxStoreAsForm(pl: PLEX; n: CARDINAL);
  (* --- parse box store, calculate parts array and window size *)
    VAR
      i0, i1, posLB, partNo, rowNo, colNo, maxColNo, picLen: CARDINAL;
      addr: SYSTEM.ADDRESS;
      outItem: BOOLEAN;
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          (* ----------------- init currents *)
          outItem := TRUE;
          picLen := 0;
          maxColNo := 0;
          partNo := 0;
          rowNo := 0;
          colNo := 0;
          i0 := 0;
          (* -------------------------- main loop *)
          WHILE i0 < storeLen DO
            posLB := min(storeLen, SrvLex.PosLineBreak(storePtr^, i0));
            (* ------------------------------------ in-row loop *)
            WHILE i0 < posLB DO
              (* ------------------------ set the beginning address *)
              addr := SYSTEM.ADR(storePtr^[i0]);
              (* ------------------------------ check the out-item glag *)
              IF outItem THEN
                (* -------------------- out of item: left separator awaited *)
                i1 := i0 + SrvLow.ScanEqInc(addr, posLB - i0, LS);
                (* ---------------------------------------- fix part if it exist *)
                IF (i0 < i1) AND (partNo < MaxNmbOfParts) THEN
                  FixPart(part[partNo], rowNo, colNo, i0, i1 - i0, picLen, outItem);
                  INC(partNo);
                END;
                (* --------------------------------------------- change state iff need *)
                outItem := (i1 >= posLB);
              ELSE
                (* --------------------------------- in the item: right separator awaited *)
                i1 := i0 + SrvLow.ScanEqInc(addr, posLB - i0, RS) + 1;
                picLen := SrvLow.ScanNeDec(SYSTEM.ADR(storePtr^[i1 - 2]), i1 - i0 - 2, PC);
                (* ------------------------------------------------ fix part if it exist *)
                IF partNo < MaxNmbOfParts THEN
                  FixPart(part[partNo], rowNo, colNo, i0, i1 - i0, picLen, outItem);
                  INC(partNo);
                END;
                (* ---------------------------------- change state iff need *)
                outItem := (i1 >= posLB) OR (storePtr^[i1 - 1] = RS);
              END;
              (* ---------------------------- increment the column *)
              INC(colNo, i1 - i0);
              (* ---------------------- shift position *)
              i0 := i1;
            END;
            (* ------------------------------- go to the next row *)
            outItem := TRUE;
            picLen := 0;
            INC(rowNo);
            IF maxColNo < colNo THEN maxColNo := colNo END;
            colNo := 0;
            i0 := posLB + 2;
          END;
          nmbOfParts := partNo;
          (* ---------------------- compute the window sizes *)
          textWdt := maxColNo;
          textHgt := rowNo;
        END;
      END;
    END PrepareBoxStoreAsForm;

  PROCEDURE PrepareBoxWindowAsForm(pl: PLEX; n: CARDINAL);
  (* --- prepare the window parameters for form *)
    VAR
      partNo, j: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        (* -------------------- replace separators by spaces *)
        FOR partNo := 0 TO nmbOfParts - 1 DO
          WITH part[partNo] DO
            IF NOT txt THEN
              (* ------------------------------------ left separator *)
              storePtr^[inx] := SP;
              (* --------------- picture characters and right separator *)
              FOR j := inx + len - pic - 1 TO inx + len - 1 DO
                storePtr^[j] := SP;
              END;
            END;
          END;
        END;
      END;
    END PrepareBoxWindowAsForm;

  PROCEDURE DrawForm(pl: PLEX; n: CARDINAL);
  (* --- draw the form according to its parameters *)
    CONST
      MaxPartLen = 200;
    TYPE
      PartString = ARRAY [0 .. MaxPartLen - 1] OF CHAR;
    VAR
      bp: Geo.Point;
      ps: PartString;
      j, partLen: CARDINAL;
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          FOR j := 0 TO nmbOfParts - 1 DO
            WITH part[j] DO
              IF len > 0 THEN
                bp.ix := attrs.leftGap + col * charWdt;
                bp.iy := attrs.topGap  + row * charHgt;
                partLen := min(len, MaxPartLen);
                SrvLow.MoveInc(SrvSys.Add(storePtr, inx), SYSTEM.ADR(ps), partLen);
                IF partLen < MaxPartLen THEN ps[partLen] := 0C END;
                IF txt THEN
                  WinOut.String(win, textAttrs, bp, ps);
                ELSE
                  WinOut.String(win, itemAttrs, bp, ps);
                END;
              END;
            END;
          END;
        END;
      END;
    END DrawForm;

  (* ========================================= Running line utilities *)

  PROCEDURE PrepareBoxStoreAsRunLine(pl: PLEX; n: CARDINAL);
  (* --- prepare box store for using it as running line *)
    VAR
      j, scrWdt, scrHgt, maxCharsNmb: CARDINAL;
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          (* -------------- replace controls by spaces *)
          FOR j := 0 TO storeLen - 1 DO
            IF storePtr^[j] < SP THEN storePtr^[j] := SP END;
          END;
          (* ------------------------ compute window size in characters *)
          Scr.Dims(scrWdt, scrHgt);
          maxCharsNmb := scrWdt DIV charWdt;
          IF (attrs.runCharsNmb = 0) OR (attrs.runCharsNmb > maxCharsNmb) THEN
            textWdt := maxCharsNmb;
          END;
          textHgt := 1;
        END;
      END;
    END PrepareBoxStoreAsRunLine;

  PROCEDURE PrepareBoxWindowAsRunLine(pl: PLEX; n: CARDINAL);
  (* --- prepare the window parameters for running line window *)
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          runIndex := 0;
          runDstPt.ix := attrs.leftGap;
          runDstPt.iy := attrs.topGap;
          runRect.ixMin := attrs.leftGap;
          runRect.ixMax := wdt - attrs.rightGap - 1;
          runRect.iyMin := attrs.topGap;
          runRect.iyMax := hgt - attrs.bottomGap - 1;
        END;
      END;
    END PrepareBoxWindowAsRunLine;

  PROCEDURE DrawRunLine(pl: PLEX; n: CARDINAL);
  (* --- draw the running line according to its parameters *)
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          LOOP
            (* ------------------------------------- draw the character *)
            WinOut.Char(win, textAttrs, runDstPt, storePtr^[runIndex]);
            INC(runDstPt.ix, charWdt);
            (* ------------------------------- change the point and the index *)
            IF runDstPt.ix < VAL(INTEGER, wdt) THEN
              runIndex := (runIndex + 1) MOD storeLen;
            ELSIF runDstPt.ix = VAL(INTEGER, wdt) THEN
              runIndex := (runIndex + 1) MOD storeLen;
              DEC(runDstPt.ix, attrs.runStep);
              EXIT;
            ELSE
              DEC(runDstPt.ix, charWdt + attrs.runStep);
              EXIT;
            END;
          END;
        END;
      END;
    END DrawRunLine;

  PROCEDURE ResetPosition(pl: PLEX; n: CARDINAL);
  (* --- reset the position to draw running line after hiding *)
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          WHILE VAL(CARDINAL, runDstPt.ix) >= attrs.leftGap + charWdt DO
            IF runIndex > 0 THEN DEC(runIndex) ELSE runIndex := storeLen - 1 END;
            DEC(runDstPt.ix, charWdt);
          END;
        END;
      END;
    END ResetPosition;

  (* ===================================== Type-independent procedure layer *)

  PROCEDURE PrepareBoxStore(pl: PLEX; n: CARDINAL);
  (* --- prepare the box store and calculate part parameters *)
    BEGIN
      IF pl^.boxPtr[n]^.type IN BoxTypeSet{vertical, horizontal, line, form} THEN
        PrepareBoxStoreAsForm(pl, n);
      ELSE
        PrepareBoxStoreAsRunLine(pl, n);
      END;
    END PrepareBoxStore;

  PROCEDURE PrepareBoxWindow(pl: PLEX; n: CARDINAL);
  (* --- prepare the window parameters *)
    BEGIN
      IF pl^.boxPtr[n]^.type IN BoxTypeSet{vertical, horizontal, line, form} THEN
        PrepareBoxWindowAsForm(pl, n);
      ELSE
        PrepareBoxWindowAsRunLine(pl, n);
      END;
    END PrepareBoxWindow;

  PROCEDURE Draw(pl: PLEX; n: CARDINAL);
  (* --- draw the box content *)
    BEGIN
      IF pl^.boxPtr[n]^.type IN BoxTypeSet{vertical, horizontal, line, form} THEN
        DrawForm(pl, n);
      ELSE
        DrawRunLine(pl, n);
      END;
    END Draw;

  (* ============================== Lexical analysis of the box description *)

  PROCEDURE ExtractHeader(VAR s: ARRAY OF CHAR; min: CARDINAL;
                          VAR t: BoxType; VAR p: BoxPosition; VAR n: CARDINAL): BOOLEAN;
  (* --- extract parameters of box header: type, position and number *)
    CONST
      Allowed = SrvLex.CharSet{"A" .. "z"};
      HeaderMark = "BOX";
    VAR
      ms: MarkString;
      bt: BoxTypeString;
      bp: BoxPosString;
      j: CARDINAL;
    BEGIN
      (* ----------------------------------- extract header mark *)
      j := SrvLex.Extract(s, min + 1, Allowed, ms);
      Caps(ms);
      IF Strings.CompareStr(ms, HeaderMark) # 0 THEN
        RETURN FALSE;
      END;
      (* --------------------------------- extract and search box type *)
      j := SrvLex.Extract(s, SrvLex.Skip(s, j), Allowed, bt);
      Caps(bt);
      t := MIN(BoxType);
      LOOP
        IF Strings.CompareStr(bt, BT[t]) = 0 THEN
          EXIT;
        ELSIF t < MAX(BoxType) THEN
          INC(t);
        ELSE
          RETURN FALSE;
        END;
      END;
      (* --------------------------------- extract and search box position *)
      j := SrvLex.Extract(s, SrvLex.Skip(s, j), Allowed, bp);
      Caps(bp);
      p := MIN(BoxPosition);
      LOOP
        IF Strings.CompareStr(bp, BP[p]) = 0 THEN
          EXIT;
        ELSIF p < MAX(BoxPosition) THEN
          INC(p);
        ELSE
          RETURN FALSE;
        END;
      END;
      (* --------------------------- extract box number *)
      j := SrvLex.Skip(s, j);
      RETURN (j < SrvLex.ExtractNumber(s, j, n));
    END ExtractHeader;

  PROCEDURE ExtractItem(VAR s: ARRAY OF CHAR; min, max: CARDINAL;
                        VAR item: ItemString; VAR t: ItemType; VAR n: CARDINAL): BOOLEAN;
  (* --- extract parameters of box item: item name, item type and number *)
    CONST
      Allowed = SrvLex.CharSet{"A" .. "z"};
      ItemMark = "ITEM";
    VAR
      ms: MarkString;
      it: ItemTypeString;
      j, itemLen: CARDINAL;
    BEGIN
      (* ----------------------------------- extract item mark *)
      j := SrvLex.Extract(s, min + 1, Allowed, ms);
      Caps(ms);
      IF Strings.CompareStr(ms, ItemMark) # 0 THEN
        RETURN FALSE;
      END;
      (* ----------------------------------------- extract item string *)
      j := SrvLex.Skip(s, j);
      IF s[j] = LS THEN
        itemLen := SrvLow.ScanEqInc(SYSTEM.ADR(s[j]), max - j, RS);
        IF s[j + itemLen] = RS THEN
          IF itemLen >= MaxItemStrLen THEN itemLen := MaxItemStrLen - 1 END;
          SrvLow.MoveInc(SYSTEM.ADR(s[j]), SYSTEM.ADR(item[0]), itemLen + 1);
          IF itemLen < MaxItemStrLen - 1 THEN item[itemLen + 1] := 0C END;
        ELSE
          RETURN FALSE;
        END;
      ELSE
        RETURN FALSE;
      END;
      (* ------------------------------------ extract and search item type *)
      j := SrvLex.Extract(s, SrvLex.Skip(s, j + itemLen + 1), Allowed, it);
      Caps(it);
      t := MIN(ItemType);
      LOOP
        IF Strings.CompareStr(it, IT[t]) = 0 THEN
          EXIT;
        ELSIF t < MAX(ItemType) THEN
          INC(t);
        ELSE
          RETURN FALSE;
        END;
      END;
      (* ------------------------------- extract number *)
      j := SrvLex.Skip(s, j);
      RETURN (j < SrvLex.ExtractNumber(s, j, n));
    END ExtractItem;

  PROCEDURE ExtractEnd(VAR s: ARRAY OF CHAR; min: CARDINAL): BOOLEAN;
  (* --- extract end line of the box *)
    CONST
      Allowed = SrvLex.CharSet{"A" .. "z"};
      EndMark = "END";
    VAR
      ms: MarkString;
      j: CARDINAL;
    BEGIN
      j := SrvLex.Extract(s, min + 1, Allowed, ms);
      Caps(ms);
      RETURN Strings.CompareStr(ms, EndMark) = 0;
    END ExtractEnd;

  (* ============================================ Procedures upon plexes *)

  PROCEDURE New(VAR s: ARRAY OF CHAR; a: Attrs;
            VAR e: ErrParams; VAR pl: PLEX): Res.Status;
    CONST
      Control = "#";
      ControlLineMinLen = 4;
    VAR
      newPlex: PLEX;
      boxT: BoxType;
      boxP: BoxPosition;
      itemT: ItemType;
      itemS: ItemString;
      textS: BoxStoreString;
      srcLen, i0, i1, textL: CARDINAL;
      boxCount, boxNo, itemCount, partNo, itemK: CARDINAL;
      outBox: BOOLEAN;
      res: Res.Status;
    BEGIN
      IF NewPlex(a, newPlex) THEN
        (* ---------------------- set currents *)
        srcLen := Strings.Length(s);
        outBox := TRUE;
        boxCount := 0;
        i0 := 0;
        (* --------------------------------- main loop *)
        LOOP
          (* ----------------------------- end of description? *)
          IF i0 <= srcLen - ControlLineMinLen THEN
            (* --------------------- process the line: find it's end *)
            i1 := SrvLex.PosLineBreak(s, i0);
            (* ---------------------------------- is it non-empty line? *)
            IF (i1 > i0) AND (i1 > SrvLex.Skip(s, i0)) THEN
              (* -------------- line is non-empy: current pos out the box? *)
              IF outBox THEN
                (* --------- current pos is out the box: box header awaited *)
                IF
                  (s[i0] = Control) AND
                  ExtractHeader(s, i0, boxT, boxP, boxNo)
                THEN
                  (* ---------- header is successfully get: check box number *)
                  IF boxNo < MaxNmbOfBoxes THEN
                    (* ---------------- box number is OK: check box existance *)
                    IF NOT BoxExist(newPlex, boxNo) THEN
                      (* ------------------ no such box exists: create new box *)
                      IF NewBox(newPlex, boxNo, boxT, boxP) THEN
                        (* ------------------------- box is successfully created *)
                        IF boxT IN BoxTypeSet{vertical, horizontal, line} THEN
                          (* -------------------- initialize the empty box store *)
                          textS := "";
                          textL := 0;
                        ELSIF boxT IN BoxTypeSet{form, runLine} THEN
                          (* --------------------- locate box store from source line *)
                          i0 := i1 + 2;
                          i1 := i0 + SrvLow.ScanEqInc(SYSTEM.ADR(s[i0]), srcLen - i0, Control) - 2;
                          IF (i1 > i0) AND (i1 = SrvLex.PosLineBreak(s, i1)) THEN
                            (* --------------- box store is located: copy it to box description *)
                            IF NewBoxStore(newPlex, boxNo, SYSTEM.ADR(s[i0]), i1 - i0 + 2) THEN
                              (* ------------------- box store is copied: prepare it for usage *)
                              PrepareBoxStore(newPlex, boxNo);
                            ELSE
                              (* ---------- fail due to memory limitations *)
                              res := Res.noMemory;
                              EXIT;
                            END;
                          ELSE
                            (* ----------- fail due to invalid form info *)
                            res := Res.invalidForm;
                            EXIT;
                          END;
                        END;
                        (* --------------- reset box parameters *)
                        outBox := FALSE;
                        itemCount := 0;
                      ELSE
                        (* ------------ fail due to memory limitations *)
                        res := Res.noMemory;
                        EXIT;
                      END;
                    ELSE
                      (* ------------ fail due to doubling box number *)
                      res := Res.doubleBoxNumber;
                      EXIT;
                    END;
                  ELSE
                    (* ------------ fail due to large box number *)
                    res := Res.boxNumberTooLarge;
                    EXIT;
                  END;
                ELSE
                  (* ------------ fail during get header info *)
                  res := Res.invalidHeader;
                  EXIT;
                END;
              ELSIF
                (s[i0] = Control) AND
                ExtractEnd(s, i0)
              THEN
                (* -------------------------- end of box: check its type *)
                IF boxT IN BoxTypeSet{vertical, horizontal, line} THEN
                  (* ------------------------------------- check items counter *)
                  IF itemCount > 0 THEN
                    (* ----------------------------------- copy box store to the box *)
                    IF NewBoxStore(newPlex, boxNo, SYSTEM.ADR(textS[0]), textL) THEN
                      (* -------- box store is copied: prepare it for usage as form *)
                      PrepareBoxStoreAsForm(newPlex, boxNo);
                    ELSE
                      (* ---------------- fail due to memory limitations *)
                      res := Res.noMemory;
                      EXIT;
                    END;
                  ELSE
                    res := Res.noItems;
                    EXIT;
                  END;
                END;
                (* ------------------------- prepare the box window *)
                SetBoxExtent(newPlex, boxNo);
                PrepareBoxWindow(newPlex, boxNo);
                (* ----------------- box is successfully ended! *)
                INC(boxCount);
                outBox := TRUE;
              ELSIF
                (s[i0] = Control) AND
                ExtractItem(s, i0, i1, itemS, itemT, itemK)
              THEN
                (* --------------------------------- item line is found *)
                IF boxT IN BoxTypeSet{vertical, horizontal, line} THEN
                  (* ------------------------------ add item to box store *)
                  IF (textL > 0) AND (boxT = vertical) THEN
                    Strings.Concat(textS, LB, textS);
                  END;
                  Strings.Concat(textS, itemS, textS);
                  textL := Strings.Length(textS);
                  (* ----------------------------- attempt to add item to box *)
                  IF AddItem(newPlex, boxNo, itemCount, itemK, itemT) THEN
                    INC(itemCount);
                  ELSE
                    res := Res.tooManyItems;
                    EXIT;
                  END;
                ELSIF boxT = form THEN
                  (* -------------------- find item part in box store *)
                  IF FindItem(newPlex, boxNo, itemS, partNo) THEN
                    (* -------------------------- attempt to add item to box *)
                    IF AddItem(newPlex, boxNo, partNo, itemK, itemT) THEN
                      (* ------------------------------------ OK: no work *)
                    ELSE
                      (* ------------------ too many items in the box *)
                      res := Res.tooManyItems;
                      EXIT;
                    END;
                  ELSE
                    (* -------- corresponding item isn't found *)
                    res := Res.itemNotFound;
                    EXIT;
                  END;
                ELSIF boxT = runLine THEN
                  (* -------------------- all the items are ignored *)
                END;
              ELSE
                (* ------------ fail during extraction of the item line *)
                res := Res.invalidItemLine;
                EXIT;
              END;
            END;
            (* ------------------- skip the line break *)
            i0 := i1 + 2;
          ELSIF NOT outBox THEN
            (* ----------------- description not complete *)
            res := Res.descriptionNotComplete;
            EXIT;
          ELSIF boxCount = 0 THEN
            (* ---------------------- empty description *)
            res := Res.descriptionEmpty;
            EXIT;
          ELSE
            (* ---------- normal exit *)
            res := Res.done;
            EXIT;
          END;
        END;
        (* ----------------------- errors processing *)
        IF res = Res.done THEN
          (* --------------- include the plex into the plex chain *)
          newPlex^.next := topPlex;
          topPlex := newPlex;
          pl := newPlex;
        ELSE
          (* -------------------------- dispose all the created boxes *)
          DisposeAllBoxes(newPlex);
          (* -------------------------- dispose the plex descriptor *)
          DrvMem.Free(newPlex, SIZE(PlexDescriptor));
          (* -------------------- report position of the error *)
          e.inx := i0;
          e.len := i1 - i0;
        END;
        (* ------------- report result *)
        RETURN res;
      ELSE
        (* -------- fail due to memory limitations *)
        RETURN Res.noMemory;
      END;
    END New;

  PROCEDURE Dispose(VAR pl: PLEX);
    VAR
      prevPlex: PLEX;
    BEGIN
      prevPlex := PLEX(SrvObj.Prev(SrvObj.OBJECT(topPlex), SrvObj.OBJECT(pl)));
      IF prevPlex # NIL THEN
        DisposeAllBoxes(pl);
        (* ------------------ exclude the plex from the plex chain *)
        IF pl # topPlex THEN
          prevPlex^.next := pl^.next;
        ELSE
          topPlex := pl^.next;
        END;
        (* --------- dispose the plex descriptor from the memory *)
        DrvMem.Free(pl, SIZE(PlexDescriptor));
      END;
    END Dispose;

  PROCEDURE Exist(pl: PLEX): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(topPlex), SrvObj.OBJECT(pl));
    END Exist;

  PROCEDURE Attributes(pl: PLEX; VAR a: Attrs);
    BEGIN
      a := pl^.attrs;
    END Attributes;

  (* =========================================== Procedures upon boxes *)

  PROCEDURE Show(pl: PLEX; n: CARDINAL): Res.Status;
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          IF winExist THEN
            (* --------- window exists: put it on the top *)
            IF Win.PutOnTop(win) = Res.done THEN
              IF type = runLine THEN
                WinOut.Roll(win, runRect, - VAL(INTEGER, attrs.runStep), 0);
                DrawRunLine(pl, n);
                RETURN Win.Update(win);
              ELSE
                RETURN Res.done;
              END;
            ELSE
              RETURN Res.noMemory;
            END;
          ELSE
            (* ----------------- create the window: it will be on top *)
            IF NewBoxWindow(pl, n) THEN
              Draw(pl, n);
              winExist := TRUE;
              RETURN Win.Show(win);
            ELSE
              RETURN Res.noMemory;
            END;
          END;
        END;
      END;
    END Show;

  PROCEDURE Hide(pl: PLEX; n: CARDINAL): Res.Status;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF winExist THEN
          (* ------------------------ dispose the graphics window *)
          IF Win.Dispose(win) = Res.done THEN
            IF type = runLine THEN ResetPosition(pl, n) END;
            winExist := FALSE;
            RETURN Res.done;
          ELSE
            RETURN Res.noMemory;
          END;
        ELSE
          (* ---------- window doesn't exist *)
          RETURN Res.done;
        END;
      END;
    END Hide;

  PROCEDURE Move(pl: PLEX; n: CARDINAL; p: Geo.Point): Res.Status;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF position = movable THEN
          IF winExist THEN
            (* -------------------- move the graphics window *)
            RETURN Win.Move(win, p);
          ELSE
            (* ------------------------ set the base point *)
            basePt := p;
            RETURN Res.done;
          END;
        ELSE
          RETURN Res.done;
        END;
      END;
    END Move;

  PROCEDURE WinExist(pl: PLEX; n: CARDINAL): BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN winExist;
      END;
    END WinExist;

  PROCEDURE Window(pl: PLEX; n: CARDINAL): Win.WINDOW;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN win;
      END;
    END Window;

  PROCEDURE Type(pl: PLEX; n: CARDINAL): BoxType;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN type;
      END;
    END Type;

  PROCEDURE Position(pl: PLEX; n: CARDINAL): BoxPosition;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN position;
      END;
    END Position;

  PROCEDURE BasePt(pl: PLEX; n: CARDINAL; VAR p: Geo.Point);
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        p := basePt;
      END;
    END BasePt;

  PROCEDURE BoxExist(pl: PLEX; n: CARDINAL): BOOLEAN;
    BEGIN
      RETURN (n < MaxNmbOfBoxes) AND (pl^.boxPtr[n] # NIL);
    END BoxExist;

  (* ========================================== Procedures upon items *)

  PROCEDURE ItemsNmb(pl: PLEX; n: CARDINAL; s: ItemTypeSet): CARDINAL;
    VAR
      count, j: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          count := 0;
          FOR j := 0 TO nmbOfItems - 1 DO
            IF item[j].type IN s THEN
              INC(count);
            END;
          END;
          RETURN count;
        ELSE
          RETURN 0;
        END;
      END;
    END ItemsNmb;

  PROCEDURE NmbOfItem(pl: PLEX; n: CARDINAL): CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN itemCurrent;
      END;
    END NmbOfItem;

  PROCEDURE TypeOfItem(pl: PLEX; n: CARDINAL): ItemType;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN item[itemCurrent].type;
      END;
    END TypeOfItem;

  PROCEDURE KeyOfItem(pl: PLEX; n: CARDINAL): CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN item[itemCurrent].key;
      END;
    END KeyOfItem;

  PROCEDURE RectOfItem(pl: PLEX; n: CARDINAL; VAR r: Geo.Rectangle);
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          WITH part[item[itemCurrent].partNo] DO
            (* --------------------------------------- x-coordinates *)
            IF type = vertical THEN
              r.ixMin := 0;
              r.ixMax := wdt - 1;
            ELSE
              r.ixMin := attrs.leftGap + col * charWdt;
              r.ixMax := r.ixMin + VAL(INTEGER, len * charWdt) - 1;
            END;
            (* ---------------------------------------------- y-coordinates *)
            IF type IN BoxTypeSet{horizontal, line} THEN
              r.iyMin := 0;
              r.iyMax := hgt - 1;
            ELSE
              r.iyMin := attrs.topGap + row * charHgt;
              r.iyMax := r.iyMin + VAL(INTEGER, charHgt) - 1;
            END;
          END;
        END;
      END;
    END RectOfItem;

  PROCEDURE PicRectOfItemNotEmpty(pl: PLEX; n: CARDINAL): BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN part[item[itemCurrent].partNo].pic > 0;
      END;
    END PicRectOfItemNotEmpty;

  PROCEDURE PicRectOfItem(pl: PLEX; n: CARDINAL; VAR r: Geo.Rectangle);
    BEGIN
      WITH pl^ DO
        WITH boxPtr[n]^ DO
          WITH part[item[itemCurrent].partNo] DO
            (* ------------------------------------------ x-coordinates *)
            r.ixMin := attrs.leftGap + (col + len - pic - 1) * charWdt;
            r.ixMax := r.ixMin + VAL(INTEGER, pic * charWdt) - 1;
            (* --------------------------------------------- y-coordinates *)
            IF type IN BoxTypeSet{horizontal, line} THEN
              r.iyMin := 0;
              r.iyMax := hgt - 1;
            ELSE
              r.iyMin := attrs.topGap + row * charHgt;
              r.iyMax := r.iyMin + VAL(INTEGER, charHgt) - 1;
            END;
          END;
        END;
      END;
    END PicRectOfItem;

  PROCEDURE SetFirstItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      j: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          FOR j := 0 TO nmbOfItems - 1 DO
            IF item[j].type IN s THEN
              itemCurrent := j;
              RETURN;
            END;
          END;
        END;
      END;
    END SetFirstItem;

  PROCEDURE SetLastItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      j: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          FOR j := nmbOfItems - 1 TO 0 BY - 1 DO
            IF item[j].type IN s THEN
              itemCurrent := j;
              RETURN;
            END;
          END;
        END;
      END;
    END SetLastItem;

  PROCEDURE SetNextItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      j: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          (* -------------------------------- up to the end *)
          FOR j := itemCurrent + 1 TO nmbOfItems - 1 DO
            IF item[j].type IN s THEN
              itemCurrent := j;
              RETURN;
            END;
          END;
          (* ---------------------------- and from the beginning *)
          FOR j := 0 TO itemCurrent DO
            IF item[j].type IN s THEN
              itemCurrent := j;
              RETURN;
            END;
          END;
        END;
      END;
    END SetNextItem;

  PROCEDURE SetPrevItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      j: CARDINAL;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          (* -------------------------------- up to the beginning *)
          IF itemCurrent > 0 THEN
            FOR j := itemCurrent - 1 TO 0 BY - 1 DO
              IF item[j].type IN s THEN
                itemCurrent := j;
                RETURN;
              END;
            END;
          END;
          (* ---------------------------- and from the end *)
          FOR j := nmbOfItems - 1 TO 0 BY - 1 DO
            IF item[j].type IN s THEN
              itemCurrent := j;
              RETURN;
            END;
          END;
        END;
      END;
    END SetPrevItem;

  PROCEDURE SetUpItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      row0, rowMax, col0, colAbs, abs, j, itemFound: CARDINAL;
      found: BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          (* ------------------------ get initial coordinates *)
          WITH part[item[itemCurrent].partNo] DO
            row0 := row;
            col0 := col;
          END;
          (* -------------------------------------------- search loop *)
          found := FALSE;
          FOR j := 0 TO nmbOfItems - 1 DO
            WITH item[j] DO
              IF type IN s THEN
                WITH part[partNo] DO
                  IF row < row0 THEN
                    IF NOT found THEN
                      rowMax := row;
                      colAbs := ABS(VAL(INTEGER, col) - VAL(INTEGER, col0));
                      itemFound := j;
                      found := TRUE;
                    ELSIF row > rowMax THEN
                      rowMax := row;
                      colAbs := ABS(VAL(INTEGER, col) - VAL(INTEGER, col0));
                      itemFound := j;
                    ELSIF row = rowMax THEN
                      abs := ABS(VAL(INTEGER, col) - VAL(INTEGER, col0));
                      IF abs < colAbs THEN
                        colAbs := abs;
                        itemFound := j;
                      END;
                    END;
                  END;
                END;
              END;
            END;
          END;
          (* ----------------- change current item, iff found *)
          IF found THEN itemCurrent := itemFound END;
        END;
      END;
    END SetUpItem;

  PROCEDURE SetDownItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      row0, rowMin, col0, colAbs, abs, j, itemFound: CARDINAL;
      found: BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          (* ------------------------ get initial coordinates *)
          WITH part[item[itemCurrent].partNo] DO
            row0 := row;
            col0 := col;
          END;
          (* -------------------------------------------- search loop *)
          found := FALSE;
          FOR j := 0 TO nmbOfItems - 1 DO
            WITH item[j] DO
              IF type IN s THEN
                WITH part[partNo] DO
                  IF row > row0 THEN
                    IF NOT found THEN
                      rowMin := row;
                      colAbs := ABS(VAL(INTEGER, col) - VAL(INTEGER, col0));
                      itemFound := j;
                      found := TRUE;
                    ELSIF row < rowMin THEN
                      rowMin := row;
                      colAbs := ABS(VAL(INTEGER, col) - VAL(INTEGER, col0));
                      itemFound := j;
                    ELSIF row = rowMin THEN
                      abs := ABS(VAL(INTEGER, col) - VAL(INTEGER, col0));
                      IF abs < colAbs THEN
                        colAbs := abs;
                        itemFound := j;
                      END;
                    END;
                  END;
                END;
              END;
            END;
          END;
          (* ----------------- change current item, iff found *)
          IF found THEN itemCurrent := itemFound END;
        END;
      END;
    END SetDownItem;

  PROCEDURE SetLeftItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      row0, col0, colMax, j, itemFound: CARDINAL;
      found: BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          (* ----------------------- get initial coordinates *)
          WITH part[item[itemCurrent].partNo] DO
            row0 := row;
            col0 := col;
          END;
          (* -------------------------------------------- search loop *)
          found := FALSE;
          FOR j := 0 TO nmbOfItems - 1 DO
            WITH item[j] DO
              IF type IN s THEN
                WITH part[partNo] DO
                  IF (row = row0) AND (col < col0) THEN
                    IF NOT found THEN
                      colMax := col;
                      itemFound := j;
                      found := TRUE;
                    ELSIF col > colMax THEN
                      colMax := col;
                      itemFound := j;
                    END;
                  END;
                END;
              END;
            END;
          END;
          (* ----------------- change current item, iff found *)
          IF found THEN itemCurrent := itemFound END;
        END;
      END;
    END SetLeftItem;

  PROCEDURE SetRightItem(pl: PLEX; n: CARDINAL; s: ItemTypeSet);
    VAR
      row0, col0, colMin, j, itemFound: CARDINAL;
      found: BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          (* ---------------------------- get initial coordinates *)
          WITH part[item[itemCurrent].partNo] DO
            row0 := row;
            col0 := col;
          END;
          (* -------------------------------------------- search loop *)
          found := FALSE;
          FOR j := 0 TO nmbOfItems - 1 DO
            WITH item[j] DO
              IF type IN s THEN
                WITH part[item[j].partNo] DO
                  IF (row = row0) AND (col > col0) THEN
                    IF NOT found THEN
                      colMin := col;
                      itemFound := j;
                      found := TRUE;
                    ELSIF col < colMin THEN
                      colMin := col;
                      itemFound := j;
                    END;
                  END;
                END;
              END;
            END;
          END;
          (* ----------------- change current item, iff found *)
          IF found THEN itemCurrent := itemFound END;
        END;
      END;
    END SetRightItem;

  PROCEDURE SetItemByNmb(pl: PLEX; n, nmb: CARDINAL);
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        IF nmbOfItems > 0 THEN
          itemCurrent := min(nmb, nmbOfItems - 1);
        END;
      END;
    END SetItemByNmb;

  PROCEDURE MarkItem(pl: PLEX; n: CARDINAL);
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        item[itemCurrent].mark := TRUE;
      END;
    END MarkItem;

  PROCEDURE UnmarkItem(pl: PLEX; n: CARDINAL);
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        item[itemCurrent].mark := FALSE;
      END;
    END UnmarkItem;

  PROCEDURE ItemMarked(pl: PLEX; n: CARDINAL): BOOLEAN;
    BEGIN
      WITH pl^.boxPtr[n]^ DO
        RETURN item[itemCurrent].mark;
      END;
    END ItemMarked;

  (* ========================= Initialization *)

  BEGIN
    topPlex := NIL;
  END Dlg.
