IMPLEMENTATION MODULE DlgInp;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *********************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    ---------------------------------------------------------------------
    A.Iakovlev  Ver  1.0                                      July 89
    A.Iakovlev  Ver  2.0  Menu attributes                     July 89
    A.Iakovlev  Ver  3.0  Parameter hide in ReadMenuSet       July 89
    A.Iakovlev  Ver  3.1  Use module Snd                      July 89
    A.Iakovlev  Ver  4.0  Background procedure                August 89
    A.Iakovlev  Ver  4.1  Use M2VERS utility                  August 89
    A.Iakovlev  Ver  5.0  Renaming                            January 90
    A.Iakovlev  Ver  5.1  New module DrvKbd                   July 90
    A.Iakovlev  Ver  6.0  Full redesign                       March 91
    A.Iakovlev  Ver  7.0  Use module Res                      August 92
    A.Iakovlev  Ver  8.0  Pragma data(var_enum_size)          August 92
    A.Iakovlev  Ver  9.0  Renaming                            June 93
    A.Iakovlev  Ver  9.1  Use ScrArOut.CorrectAttrs           February 94
    A.Iakovlev  Ver  9.2  Use module Mem for heap control     October 94
    A.Iakovlev  Ver 10.0  Simplifications in attributes       March 95
    A.Iakovlev  Ver 10.1  Use new ScrOut module               March 95
    *********************************************************************
    TopSpeed compiler dependencies:
    ------ pragmas (*#...*)
  *)

  IMPORT
    Geo, Res, DrvMem, DrvKbd, DrvMou, DrvSnd,
    Scr, ScrOut, Win, WinOut, Dlg;

  (* ===================================================== Utilities *)

  PROCEDURE ShowBox(pl: Dlg.PLEX; n: CARDINAL; init: FaceProc);
    CONST
      Items = Dlg.ItemTypeSet{Dlg.view .. Dlg.callExit};
    VAR
      u: Win.WINDOW;
      oldV, curV: Scr.View;
      j, len, no: CARDINAL;
    BEGIN
      (* ------------------------------- get box info *)
      len := Dlg.ItemsNmb(pl, n, Items);
      (* ------------------------- view data via box window *)
      IF len > 0 THEN
        (* ----------------------------------- get the window *)
        u := Dlg.Window(pl, n);
        (* -------------------------- store the window viewport *)
        Win.Viewport(u, oldV);
        (* ------------------------------------------ set clip mode *)
        curV.clip := TRUE;
        (* --------------------------------- store the old item number *)
        no := Dlg.NmbOfItem(pl, n);
        (* ----------------------------------------- set the first item *)
        Dlg.SetFirstItem(pl, n, Items);
        (* ------------------------------------------------- main loop *)
        FOR j := 0 TO len - 1 DO
          (* ---------------------------- test item picture rectangle *)
          IF Dlg.PicRectOfItemNotEmpty(pl, n) THEN
            (* ------------------------- get item picture rectangle *)
            Dlg.PicRectOfItem(pl, n, curV.port);
            Win.SetViewport(u, curV);
            (* ---------------------------- call face procedure *)
            init(u, Dlg.KeyOfItem(pl, n));
          END;
          (* ------------------------- go to the next item *)
          Dlg.SetNextItem(pl, n, Items);
        END;
        (* ------------------- restore the item number *)
        Dlg.SetItemByNmb(pl, n, no);
        (* ------------- restore the window viewport *)
        Win.SetViewport(u, oldV);
      END;
    END ShowBox;

  (* ==================================================== Main procedures *)

  PROCEDURE Show(pl: Dlg.PLEX; n: CARDINAL; init: FaceProc): Res.Status;
    VAR
      u: Win.WINDOW;
    BEGIN
      IF Dlg.Exist(pl) THEN
        IF Dlg.BoxExist(pl, n) THEN
          IF Dlg.WinExist(pl, n) THEN
            ShowBox(pl, n, init);
            RETURN Res.done;
          ELSE
            RETURN Res.boxNotShown;
          END;
        ELSE
          RETURN Res.boxNotExist;
        END;
      ELSE
        RETURN Res.invalidPlex;
      END;
    END Show;

  PROCEDURE Act(pl: Dlg.PLEX; n: CARDINAL;
                VAR a: Attrs; init, call: FaceProc;
                wait: PROC; get: GetAction): Res.Status;

    TYPE
      BoxTypeSet = SET OF Dlg.BoxType;
      BoxAction = (exit, nextBox, previousBox, tempPreviousBox);
      State =
        RECORD
          boxNo: CARDINAL;
          boxPt: Geo.Point;
          boxTS: BoxTypeSet;
        END;

    CONST
      MaxStackLen = MAX(CARDINAL) DIV SIZE(State);
      (* ------------------------------------------------- *)
      EmpTypeSet = BoxTypeSet{};
      VerTypeSet = BoxTypeSet{Dlg.vertical};
      HorTypeSet = BoxTypeSet{Dlg.horizontal, Dlg.line};
      (* ----------------------------------------------------------- *)
      Items = Dlg.ItemTypeSet{Dlg.call .. Dlg.link};

    TYPE
      Stack = ARRAY [0 .. MaxStackLen - 1] OF State;
      StackPtr = POINTER TO Stack;

    VAR
      buffPtr: StackPtr;
      st, stNext: State;
      act: Action;
      inx, combShift: CARDINAL;
      forwardFlag, getFlag: BOOLEAN;
      plexRes, res: Res.Status;

    PROCEDURE GetShift(): CARDINAL;
    (* --- compute the vertical shift for horizontal/line boxes *)
      VAR
        plexAttrs: Dlg.Attrs;
        w, h: CARDINAL;
      BEGIN
        Dlg.Attributes(pl, plexAttrs);
        Win.FrameDims(plexAttrs.frameAttrs, w, h);
        RETURN h + Win.TitleHgt(plexAttrs.titleAttrs) + 1;
      END GetShift;

    PROCEDURE DoBox(): BoxAction;

      VAR
        t: Dlg.BoxType;
        i: Dlg.ItemType;
        r: Geo.Rectangle;
        new, old: Scr.View;
        cursAttrs: ScrOut.Attrs;
        u: Win.WINDOW;
        s: BoxAction;

      PROCEDURE InitCursorAttrs();
        VAR
          p: Geo.Point;
        BEGIN
          ScrOut.SetFillAttrs
          (
            cursAttrs,
            Scr.xor, a.cursColor, 0,
            ScrOut.fillSolid, ScrOut.FONT(NIL),
            p, 0
          );
        END InitCursorAttrs;

      PROCEDURE DrawCursor();
      (* --- draw rectangular cursor at the current item of current box *)
        VAR
          winRes: Res.Status;
        BEGIN
          WinOut.FillRectangle(u, cursAttrs, r);
          winRes := Win.Update(u);
        END DrawCursor;

      PROCEDURE SetNextState(VAR ns: State);
      (* --- determine the next state *)
        VAR
          p: Geo.Point;
        BEGIN
          (* ----------------------------------------- box number *)
          ns.boxNo := Dlg.KeyOfItem(pl, st.boxNo);
          (* ---------------------------------------------- base point *)
          Dlg.BasePt(pl, st.boxNo, p);
          CASE t OF
          | Dlg.vertical, Dlg.form:
            INC(p.ix, r.ixMin + VAL(INTEGER, a.horShift));
            INC(p.iy, r.iyMin + VAL(INTEGER, a.verShift));
          | Dlg.horizontal, Dlg.line:
            INC(p.ix, r.ixMin);
            INC(p.iy, r.iyMax + VAL(INTEGER, combShift));
          END;
          ns.boxPt := p;
          (* --------------------------- box types on path *)
          ns.boxTS := st.boxTS;
        END SetNextState;

      BEGIN
        (* -------------- initialize cursor attributes *)
        InitCursorAttrs();
        (* --------------------------------- check items number *)
        IF Dlg.ItemsNmb(pl, st.boxNo, Items) > 0 THEN
          (* --------------------------------------------- get box info *)
          t := Dlg.Type(pl, st.boxNo);
          INCL(st.boxTS, t);
          u := Dlg.Window(pl, st.boxNo);
          (* ------------------------------------ adjust the current item *)
          IF NOT (Dlg.TypeOfItem(pl, st.boxNo) IN Items) THEN
            Dlg.SetFirstItem(pl, st.boxNo, Items);
          END;
          (* ------------------- get the rectangle of the current item *)
          Dlg.RectOfItem(pl, st.boxNo, r);
          (* -------------------------------------- draw the cursor *)
          DrawCursor();
          (* ---------------------------------- intra-box loop *)
          LOOP
            (* ---------------------- get the action iff need *)
            IF getFlag THEN
              IF Dlg.ItemMarked(pl, st.boxNo) THEN
                act := enter;
              ELSE
                act := get(wait);
              END;
            ELSE
              getFlag := TRUE;
            END;
            (* ----------------------------- process the action *)
            CASE act OF
            | enter:
              (* -------------------------------- check the item type *)
              i := Dlg.TypeOfItem(pl, st.boxNo);
              CASE i OF
              | Dlg.call, Dlg.callExit:
                (* --------------------------------- call: hide the cursor *)
                DrawCursor();
                (* ---------------------------- get and set the window viewport *)
                new.clip := TRUE;
                Dlg.PicRectOfItem(pl, st.boxNo, new.port);
                Win.SwapViewport(u, new, old);
                (* ---------------------------------- call the procedure *)
                call(u, Dlg.KeyOfItem(pl, st.boxNo));
                (* ----------------------- restore the window viewport *)
                Win.SetViewport(u, old);
                (* ------------------------ draw the cursor again *)
                DrawCursor();
                (* ---------------------------- exit iff need *)
                IF i = Dlg.callExit THEN
                  s := exit;
                  EXIT;
                END;
              | Dlg.link:
                (* -------------------------------- set the next state *)
                SetNextState(stNext);
                (* --------------------------- exit *)
                s := nextBox;
                EXIT;
              END;
            | escape:
              (* ---------------- leave the current box *)
              s := previousBox;
              EXIT;
            | first:
              (* ------------------------ set the first item *)
              Dlg.SetFirstItem(pl, st.boxNo, Items);
            | last:
              (* ----------------------------- set the last item *)
              Dlg.SetLastItem(pl, st.boxNo, Items);
            | up:
              (* ------------------------------ set the up item *)
              CASE t OF
              | Dlg.vertical:
                Dlg.SetPrevItem(pl, st.boxNo, Items);
              | Dlg.horizontal:
                IF st.boxTS * VerTypeSet # EmpTypeSet THEN
                  getFlag := FALSE;
                  s := tempPreviousBox;
                  EXIT;
                END;
              | Dlg.form:
                Dlg.SetUpItem(pl, st.boxNo, Items);
              END;
            | down:
              (* -------------------------- set the down item *)
              CASE t OF
              | Dlg.vertical:
                Dlg.SetNextItem(pl, st.boxNo, Items);
              | Dlg.horizontal:
                IF st.boxTS * VerTypeSet # EmpTypeSet THEN
                  getFlag := FALSE;
                  s := tempPreviousBox;
                  EXIT;
                END;
              | Dlg.form:
                Dlg.SetDownItem(pl, st.boxNo, Items);
              END;
            | left:
              (* ------------------------ set the left item *)
              CASE t OF
              | Dlg.vertical:
                IF st.boxTS * HorTypeSet # EmpTypeSet THEN
                  getFlag := FALSE;
                  s := tempPreviousBox;
                  EXIT;
                END;
              | Dlg.horizontal, Dlg.line:
                Dlg.SetPrevItem(pl, st.boxNo, Items);
              | Dlg.form:
                Dlg.SetLeftItem(pl, st.boxNo, Items);
              END;
            | right:
              (* -------------------- set the right item *)
              CASE t OF
              | Dlg.vertical:
                IF st.boxTS * HorTypeSet # EmpTypeSet THEN
                  getFlag := FALSE;
                  s := tempPreviousBox;
                  EXIT;
                END;
              | Dlg.horizontal, Dlg.line:
                Dlg.SetNextItem(pl, st.boxNo, Items);
              | Dlg.form:
                Dlg.SetRightItem(pl, st.boxNo, Items);
              END;
            END;
            (* -------------------- move the cursor *)
            DrawCursor();
            Dlg.RectOfItem(pl, st.boxNo, r);
            DrawCursor();
          END;
          (* ------------------------ hide the cursor *)
          DrawCursor();
          (* ------------------- normal exit *)
          RETURN s;
        ELSE
          (* ---------------------------- box has no items: intra-box loop *)
          LOOP
            (* ------------------------------------- get action *)
            act := get(wait);
            (* ------------------------ process the action *)
            CASE act OF
            | enter, escape:
              (* ---------------- leave the current box *)
              s := previousBox;
              EXIT;
            | first, last:
              (* ------------------------ set the first item *)
              getFlag := FALSE;
              s := tempPreviousBox;
              EXIT;
            | up, down:
              (* ------------------------------ set the up item *)
              IF st.boxTS * VerTypeSet # EmpTypeSet THEN
                getFlag := FALSE;
                s := tempPreviousBox;
                EXIT;
              END;
            | left, right:
              (* ------------------------ set the left item *)
              IF st.boxTS * HorTypeSet # EmpTypeSet THEN
                getFlag := FALSE;
                s := tempPreviousBox;
                EXIT;
              END;
            END;
          END;
          (* ------------------- normal exit *)
          RETURN s;
        END;
      END DoBox;

    BEGIN
      IF Dlg.Exist(pl) THEN
        (* ----------------- correct depth and allocate the stack workspace *)
        IF a.depth > MaxStackLen THEN a.depth := MaxStackLen END;
        IF NOT DrvMem.Alloc(buffPtr, a.depth * SIZE(State)) THEN
          RETURN Res.noMemory;
        END;
        (* ------------------------ determine combined vertical shift *)
        combShift := GetShift();
        (* ---------------------------- init top state *)
        st.boxNo := n;
        Dlg.BasePt(pl, n, st.boxPt);
        st.boxTS := EmpTypeSet;
        (* ------------------------ init controls *)
        forwardFlag := TRUE;
        getFlag := TRUE;
        inx := 0;
        (* ---------------------------------------- main loop *)
        LOOP
          IF forwardFlag THEN
            (* ------------------------------- check box existence *)
            IF Dlg.BoxExist(pl, st.boxNo) THEN
              (* ------------------------ box exists: move and show it *)
              IF
                (Dlg.Move(pl, st.boxNo, st.boxPt) = Res.done) AND
                (Dlg.Show(pl, st.boxNo) = Res.done)
              THEN
                (* ----------------------- box has shown: show data via it *)
                ShowBox(pl, st.boxNo, init);
                (* -------------------------------------- interact with box *)
                CASE DoBox() OF
                | exit:
                  (* ------------------------------ hide the current box *)
                  plexRes := Dlg.Hide(pl, st.boxNo);
                  (* ---------------------- exit with result = done *)
                  forwardFlag := FALSE;
                  res := Res.done;
                | nextBox:
                  (* ----------------- check the depth of the stack *)
                  IF inx < a.depth THEN
                    (* ------------------ hide the current box iff need *)
                    IF a.hide THEN
                      plexRes := Dlg.Hide(pl, st.boxNo);
                    END;
                    (* ------------------ store the current state *)
                    buffPtr^[inx] := st;
                    INC(inx);
                    (* --------------------- set the next state *)
                    st := stNext;
                  ELSE
                    (* ------------------- fail due to small stack *)
                    plexRes := Dlg.Hide(pl, st.boxNo);
                    forwardFlag := FALSE;
                    res := Res.depthTooSmall;
                  END;
                | previousBox:
                  (* -------------------------- hide the current box *)
                  plexRes := Dlg.Hide(pl, st.boxNo);
                  (* ----------------- go to the previous box, iff it exists *)
                  IF inx > 0 THEN
                    DEC(inx);
                    st := buffPtr^[inx];
                    Dlg.UnmarkItem(pl, st.boxNo);
                  ELSE
                    res := Res.break;
                    EXIT;
                  END;
                | tempPreviousBox:
                  (* -------------------------- hide the current box *)
                  plexRes := Dlg.Hide(pl, st.boxNo);
                  (* ---------------------------- go to the previous box *)
                  DEC(inx);
                  st := buffPtr^[inx];
                  Dlg.MarkItem(pl, st.boxNo);
                END;
              ELSE
                (* ---------- box can't be shown: exit with result = noMemory *)
                forwardFlag := FALSE;
                res := Res.noMemory;
              END;
            ELSE
              (* ----------------- no box: exit with result = invalidPlex *)
              forwardFlag := FALSE;
              res := Res.boxNotExist;
            END;
          ELSIF inx > 0 THEN
            (* --------------------- return to the previous box *)
            DEC(inx);
            st := buffPtr^[inx];
            plexRes := Dlg.Hide(pl, st.boxNo);
            Dlg.MarkItem(pl, st.boxNo);
          ELSE
            (* ---------------- no previous box: exit *)
            EXIT;
          END;
        END;
        (* ---------------------------- deallocate the stack workspace *)
        DrvMem.Free(buffPtr, a.depth * SIZE(State));
        (* ------------------------------------ normal exit *)
        RETURN res;
      ELSE
        RETURN Res.invalidPlex;
      END;
    END Act;

  (* ========================== Standard action-returning procedure *)

  PROCEDURE Get(wait: PROC): Action;
    VAR
      key: CARDINAL;
      s: DrvMou.ButtonSet;
      dx, dy: INTEGER;
    BEGIN
      LOOP
        IF DrvKbd.Get(key) THEN
          (* ----------------------------- keyboard actions *)
          CASE key OF
          | DrvKbd.Home:
            RETURN first;
          | DrvKbd.UpArrow:
            RETURN up;
          | DrvKbd.LeftArrow:
            RETURN left;
          | DrvKbd.RightArrow:
            RETURN right;
          | DrvKbd.End:
            RETURN last;
          | DrvKbd.DownArrow:
            RETURN down;
          | DrvKbd.Enter:
            RETURN enter;
          | DrvKbd.Escape:
            RETURN escape;
          ELSE
            DrvSnd.Error();
          END;
        ELSIF DrvMou.Get(s, dx, dy) THEN
          (* -------------------------------- mouse actions *)
          IF DrvMou.left IN s THEN
            RETURN enter;
          ELSIF DrvMou.right IN s THEN
            RETURN escape;
          ELSIF DrvMou.middle IN s THEN
            DrvSnd.Error();
          END;
        ELSE
          (* ------------- no actions: wait *)
          wait();
        END;
      END;
    END Get;

  END DlgInp.
