IMPLEMENTATION MODULE ScrFnt;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **********************************************************************
    Compiler:   TopSpeed / Stony Brook Modula-2
    ----------------------------------------------------------------------
    D.Maslov    Ver  1.0  Initial version                     August 91
    A.Iakovlev  Ver  2.0  Standartization                     November 91
    A.Iakovlev  Ver  3.0  New module ScrTxOut                 November 91
    A.Iakovlev  Ver  4.0  Use module Res                      June 93
    A.Iakovlev  Ver  4.1  Import from SYSTEM                  January 94
    A.Iakovlev  Ver  4.2  Use module SrvLow                   January 94
    A.Iakovlev  Ver  5.0  Use type ScrTxOut.CharSet           January 94
    A.Iakovlev  Ver  5.1  Use module Strings                  January 94
    A.Iakovlev  Ver  5.2  Use module Dsk                      February 94
    A.Iakovlev  Ver  5.3  Use module Mem for heap control     October 94
    A.Iakovlev  Ver  6.0  Use new ScrOut module               March 95
    **********************************************************************
    TopSpeed compiler dependencies:
    ------ options (*$...*)
    TopSpeed library dependencies:
    ------ import from Strings: Assign, Length, Delete, Concat
    Stony Brook library dependencies:
    ------ import from Strings: Assign, Length, Delete, Concat
  *)

  IMPORT
    SYSTEM, Strings,
    Res, DrvDsk, DrvMem, SrvLow, ScrOut;

  CONST
    NormVersion = 1;
    Empty = ScrOut.CharSet{};
    Full = ScrOut.CharSet{MIN(CHAR) .. MAX(CHAR)};
    StrLen = 64;

  TYPE
    String = ARRAY [0 .. StrLen - 1] OF CHAR;
    Header =
      RECORD
        CASE version: CARDINAL OF
        | NormVersion:
          wdt: CARDINAL;          (* --- width of character in pixels *)
          hgt: CARDINAL;          (* --- height of character in pixels *)
          set: ScrOut.CharSet;    (* --- characters in font *)
        END;
      END;

  (* ======================================= Utilities *)

  (*$N*)

  PROCEDURE ReplaceExt(fn, ext: ARRAY OF CHAR; VAR fnExt: ARRAY OF CHAR);
    CONST
      Point = ".";
      BackSlash = "\";
    VAR
      i, len: CARDINAL;
      found: BOOLEAN;
    BEGIN
      Strings.Assign(fn, fnExt);
      len := Strings.Length(fnExt);
      i := len - 1;
      LOOP
        IF fnExt[i] = Point THEN
          found := TRUE;
          EXIT;
        ELSIF (fnExt[i] = BackSlash) OR (i = 0) THEN
          found := FALSE;
          EXIT;
        ELSE
          DEC(i);
        END;
      END;
      IF found THEN Strings.Delete(fnExt, i, len - i) END;
      Strings.Concat(fnExt, Point, fnExt);
      Strings.Concat(fnExt, ext, fnExt);
    END ReplaceExt;

  (*$F*)

  (* ============================================= Load procedures *)

  PROCEDURE NewLoadPart(filename: ARRAY OF CHAR; VAR font: ScrOut.FONT;
            room: ScrOut.CharSet; VAR part: ScrOut.CharSet): Res.Status;
    VAR
      f: DrvDsk.File;
      buff: SYSTEM.ADDRESS;
      len: CARDINAL;
      h: Header;
      ch: CHAR;
      st: Res.Status;
    BEGIN
      (* ------------------------ does file exists? *)
      IF DrvDsk.Exists(filename) THEN
        (* ------------------------- success: open the file *)
        f := DrvDsk.Open(filename);
        (* ---------------------------- attempt to read header *)
        IF DrvDsk.Read(f, SYSTEM.ADR(h), SIZE(h)) = SIZE(h) THEN
          (* ------------------------------ success: parse the file *)
          CASE h.version OF
          | NormVersion:
            (* -------------- normal font file: set part for characters *)
            part := part * h.set;
            (* -------------------------- attempt to create the empty font *)
            CASE ScrOut.NewEmpty(h.wdt, h.hgt, part + room, font) OF
            | Res.done:
              (* ---------- font is created: calculate length of character *)
              len := ((h.wdt - 1) DIV 8 + 1) * h.hgt;
              (* -------------------- sizes are ok: is memory available? *)
              IF DrvMem.Alloc(buff, len) THEN
                (* ----------------------------------- read font file *)
                LOOP
                  FOR ch := MIN(CHAR) TO MAX(CHAR) DO
                    IF ch IN h.set THEN
                      IF DrvDsk.Read(f, buff, len) = len THEN
                        IF ch IN part THEN
                          ScrOut.SetStipple(font, ch, buff);
                        END;
                      ELSE
                        st := Res.invalidFontFile;
                        EXIT;
                      END;
                    END;
                  END;
                  st := Res.done;
                  EXIT;
                END;
                (* -------------------------- deallocate the buffer *)
                DrvMem.Free(buff, len);
              ELSE
                (* ---------------- no memory for character buffer *)
                st := Res.noMemory;
              END;
              (* --------------- dispose the font if errors were *)
              IF st # Res.done THEN ScrOut.Dispose(font) END;
            | Res.noMemory:
              (* ---------------- no memory to create the font *)
              st := Res.noMemory;
            | Res.badFontDescription:
              (* ----------- invalid sizes of font character *)
              st := Res.invalidFontFile;
            END;
          ELSE
            (* ----------------------- unknown font version *)
            st := Res.invalidFontFile;
          END;
        ELSE
          (* --------------------- font header isn't read *)
          st := Res.invalidFontFile;
        END;
        DrvDsk.Close(f);
      ELSE
        st := Res.fileNotFound;
      END;
      (* ----------- return *)
      RETURN st;
    END NewLoadPart;

  PROCEDURE NewLoad(filename: ARRAY OF CHAR; VAR font: ScrOut.FONT): Res.Status;
    VAR
      part: ScrOut.CharSet;
    BEGIN
      part := Full;
      RETURN NewLoadPart(filename, font, Full, part);
    END NewLoad;

  (* ================================================= Save procedures *)

  PROCEDURE SavePart(font: ScrOut.FONT; filename: ARRAY OF CHAR;
            room, part: ScrOut.CharSet): Res.Status;
    CONST
      BackExt = "BAK";
      TempExt = "$$$";
    VAR
      f: DrvDsk.File;
      tempName, backName: String;
      buff: SYSTEM.ADDRESS;
      len: CARDINAL;
      h: Header;
      ch: CHAR;
      st: Res.Status;
    BEGIN
      (* ------------------- check font existence *)
      IF ScrOut.Exist(font) THEN
        (* ---------------------- attempt to create font file *)
        ReplaceExt(filename, TempExt, tempName);
        f := DrvDsk.Create(tempName);
        IF DrvDsk.result = 0 THEN
          (* --------------------- create and write the font header *)
          h.version := NormVersion;
          ScrOut.Dims(font, h.wdt, h.hgt);
          ScrOut.Chars(font, h.set);
          part := part * h.set;
          h.set := h.set + room;
          (* --------------------------------------- is header written? *)
          IF DrvDsk.Write(f, SYSTEM.ADR(h), SIZE(h)) = SIZE(h) THEN
            (* ---------------------- yes: calculate length of character *)
            len := ((h.wdt - 1) DIV 8 + 1) * h.hgt;
            (* ---------------------------------- is memory available? *)
            IF DrvMem.Alloc(buff, len) THEN
              (* ---------------------------- write the font file *)
              LOOP
                FOR ch := MIN(CHAR) TO MAX(CHAR) DO
                  IF ch IN h.set THEN
                    IF ch IN part THEN
                      ScrOut.Stipple(font, ch, buff);
                    ELSE
                      SrvLow.FastFill(buff, len, 0);
                    END;
                    IF DrvDsk.Write(f, buff, len) # len THEN
                      st := Res.fileWriteError;
                      EXIT;
                    END;
                  END;
                END;
                st := Res.done;
                EXIT;
              END;
              (* ---------------------- deallocate the buffer *)
              DrvMem.Free(buff, len);
            ELSE
              (* -------- no memory for character buffer *)
              st := Res.noMemory;
            END;
          ELSE
            (* -------------- header isn't written *)
            st := Res.fileWriteError;
          END;
          (* ----------------------------- process files *)
          DrvDsk.Close(f);
          IF st = Res.done THEN
            IF DrvDsk.Exists(filename) THEN
              ReplaceExt(filename, BackExt, backName);
              IF DrvDsk.Exists(backName) THEN DrvDsk.Delete(backName) END;
              DrvDsk.Rename(filename, backName);
            END;
            DrvDsk.Rename(tempName, filename);
          ELSE
            DrvDsk.Delete(tempName);
          END;
        ELSE
          st := Res.fileNotCreated;
        END;
      ELSE
        st := Res.fontNotExist;
      END;
      (* ----------- return *)
      RETURN st;
    END SavePart;

  PROCEDURE Save(font: ScrOut.FONT; filename: ARRAY OF CHAR): Res.Status;
    BEGIN
      RETURN SavePart(font, filename, Empty, Full);
    END Save;

  END ScrFnt.
