IMPLEMENTATION MODULE ScrLow;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **************************************************************************
    Hardware:   EGA & VGA Graphics Adapters
    Compiler:   TopSpeed / Stony Brook Modula-2
    --------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from multiple Ctrl-modules       January 92
    A.Iakovlev  Ver  1.1  Version for TopSpeed DOS Extender        January 92
    A.Iakovlev  Ver  2.0  No procedures Active and Memory          January 92
    A.Iakovlev  Ver  3.0  Separate procedures for CGA/EGA/VGA      January 92
    A.Iakovlev  Ver  3.1  Mode 376 x 348 instead of 400 x 327      August 92
    A.Iakovlev  Ver  4.0  Some inline procedures for DOS           August 92
    A.Iakovlev  Ver  5.0  Renaming                                 June 93
    A.Iakovlev  Ver  5.1  No Super-VGA support under OS/2          June 93
    A.Iakovlev  Ver  6.0  Inline procedures aren't first           June 93
    A.Iakovlev  Ver  6.1  More careful Trident test and banking    June 93
    A.Iakovlev  Ver  7.0  Import from SYSTEM                       January 94
    A.Iakovlev  Ver  7.1  Use module Strings                       January 94
    A.Iakovlev  Ver  8.0  No use SHORTCARD type                    January 94
    A.Iakovlev  Ver  8.1  No disable interrupts during DAC access  January 94
    A.Iakovlev  Ver  9.0  Erasure of all the CGA-procedures        January 94
    A.Iakovlev  Ver 10.0  No OS/2 support                          February 94
    A.Iakovlev  Ver 10.1  Stony Brook support                      February 94
    A.Iakovlev  Ver 11.0  Procedure BytesPerShift                  March 94
    A.Iakovlev  Ver 12.0  No support for Paradise and Trident      March 94
    A.Iakovlev  Ver 13.0  Renaming                                 March 95
    A.Iakovlev  Ver 13.1  No checking in procedure BytesPerShift   April 95
    **************************************************************************
    Hardware dependencies:
    ------ EGA & VGA input/output ports
    TopSpeed compiler dependencies:
    ------ options (*$...*)
    ------ structured constants
    ------ single-byte sets
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE, Out, In
    Stony Brook compiler dependencies:
    ------ structured constants
    ------ single-byte sets
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE, OUTB, INB
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT SYSTEM;

  (*$I-,Z-,R-,S-*)

  (* =================== Conditional import *)

  (*%T CMPTS*)
  FROM SYSTEM IMPORT Out, In;
  (*%E*)

  (* =========================== Library interface *)

  (*%T CMPSB*)

  (*
    ------ This interface isn't an effective way, but it is necessary
    ------ to support two different parameter protocols...
  *)

  PROCEDURE Out(port: CARDINAL; data: SYSTEM.BYTE);
    BEGIN
      SYSTEM.OUTB(data, port);
    END Out;

  PROCEDURE In(port: CARDINAL): SYSTEM.BYTE;
    VAR
      data: SYSTEM.BYTE;
    BEGIN
      SYSTEM.INB(data, port);
      RETURN data;
    END In;

  (*%E*)

  (* ==================================== EGA/VGA registers *)

  TYPE
    BS = SET OF [0 .. 7];

  CONST
    (* ---------------- EGA/VGA Attribute Controller *)
    AttrAddr = 3C0H;
    AttrDataOut = 3C0H;
    AttrDataIn = 3C1H;
    ModeControlReg = 10H;
    ColorSelectReg = 14H;
    (* -------------- EGA/VGA Miscellaneous Output Register *)
    MiscOut = 3C2H;
    (* ----------------------------------- EGA/VGA Sequencer *)
    SeqAddr = 3C4H;
    SeqData = 3C5H;
    ResetReg = 0;
    ClockingModeReg = 1;
    ScreenOffBit = 5;
    MapMaskReg = 2;
    MemoryModeReg = 4;
    (* ------------------ Trident extension of EGA/VGA Sequencer *)
    TridentVersionReg = 0BH;
    TridentModeControl1Reg = 0EH;
    (* ----------------------------------------- VGA DAC Controller *)
    DacAddrIn = 3C7H;
    DacAddrOut = 3C8H;
    DacData = 3C9H;
    (* ---------------------------------- EGA/VGA Graphics Controller *)
    GraphAddr = 3CEH;
    GraphData = 3CFH;
    SetResetReg = 0;
    DataRotateReg = 3;
    ReadMapSelectReg = 4;
    ModeReg = 5;
    MiscReg = 6;
    BitMaskReg = 8;
    (* ---------------- Paradise extension of EGA/VGA Graphics Controller *)
    ParadiseAddressOffsetAReg = 9;
    ParadiseAddressOffsetBReg = 0AH;
    ParadiseVideoControlReg = 0EH;
    ParadiseGenPurposeReg = 0FH;
    (* --------------------------------------------- CRT Controller *)
    CRTAddr = 3D4H;
    CRTData = 3D5H;
    MaxScanLineReg = 9H;
    StartAddrHghReg = 0CH;
    StartAddrLowReg = 0DH;
    CursorLocHghReg = 0EH;
    VerticalRetraceEndReg = 11H;
    ProtectBit = 7;
    OffsetReg = 13H;
    UnderlineLocReg = 14H;
    CRTModeControlReg = 17H;
    (* -------------------- Paradise extension of CRT Controller *)
    ParadiseMiscControl2Reg = 2FH;
    (* --------------------- Input Status Register *)
    InputStatus1 = 3DAH;
    RetraceBit = 3;

  (* ========================================== Utilities *)

  PROCEDURE CheckReg(port: CARDINAL; reg, xor: SYSTEM.BYTE): BOOLEAN;
    CONST
      MaxCount = 200;
    VAR
      oldReg, newReg: SYSTEM.BYTE;
      count: CARDINAL;
    BEGIN
      Out(port, reg);
      oldReg := In(port + 1);
      Out(port + 1, SYSTEM.BYTE(BS(oldReg) / BS(xor)));
      count := 0;
      REPEAT INC(count) UNTIL count = MaxCount;
      newReg := In(port + 1);
      Out(port + 1, oldReg);
      RETURN newReg = SYSTEM.BYTE(BS(oldReg) / BS(xor));
    END CheckReg;

  (* ================================= Common control functions *)

  PROCEDURE WaitForRetrace();
    BEGIN
      IF RetraceBit IN BS(In(InputStatus1)) THEN
        REPEAT UNTIL NOT (RetraceBit IN BS(In(InputStatus1)));
      END;
      REPEAT UNTIL RetraceBit IN BS(In(InputStatus1));
    END WaitForRetrace;

  PROCEDURE SetStartAddr(s: CARDINAL);
    BEGIN
      WaitForRetrace();
      Out(CRTAddr, StartAddrHghReg);
      Out(CRTData, SYSTEM.BYTE(s DIV 100H));
      Out(CRTAddr, StartAddrLowReg);
      Out(CRTData, SYSTEM.BYTE(s));
    END SetStartAddr;

  PROCEDURE BytesPerShift(): CARDINAL;
    BEGIN
      Out(CRTAddr, OffsetReg);
      RETURN ORD(In(CRTData)) * 2;
    END BytesPerShift;

  (* ======================= EGA/VGA-dependent control functions *)

  PROCEDURE EnableVideo();
    VAR
      m: SYSTEM.BYTE;
    BEGIN
      m := In(InputStatus1);
      Out(AttrAddr, 20H);
    END EnableVideo;

  PROCEDURE DisableVideo();
    VAR
      m: SYSTEM.BYTE;
    BEGIN
      m := In(InputStatus1);
      Out(AttrAddr, 0);
    END DisableVideo;

  PROCEDURE SetWMode1();
    BEGIN
      Out(GraphAddr, ModeReg);
      Out(GraphData, 1);
    END SetWMode1;

  PROCEDURE SetWMode2();
    BEGIN
      Out(GraphAddr, ModeReg);
      Out(GraphData, 2);
    END SetWMode2;

  PROCEDURE ResetMode();
    BEGIN
      Out(GraphAddr, ModeReg);
      Out(GraphData, 0);
    END ResetMode;

  PROCEDURE SetBitMask(m: SYSTEM.BYTE);
    BEGIN
      Out(GraphAddr, BitMaskReg);
      Out(GraphData, m);
    END SetBitMask;

  PROCEDURE ResetBitMask();
    BEGIN
      Out(GraphAddr, BitMaskReg);
      Out(GraphData, 0FFH);
    END ResetBitMask;

  PROCEDURE SetReadMap(m: SYSTEM.BYTE);
    BEGIN
      Out(GraphAddr, ReadMapSelectReg);
      Out(GraphData, m);
    END SetReadMap;

  PROCEDURE SetMapMask(m: SYSTEM.BYTE);
    BEGIN
      Out(SeqAddr, MapMaskReg);
      Out(SeqData, m);
    END SetMapMask;

  PROCEDURE ResetMapMask();
    BEGIN
      Out(SeqAddr, MapMaskReg);
      Out(SeqData, 0FH);
    END ResetMapMask;

  PROCEDURE SetFunc(func: SYSTEM.BYTE);
    BEGIN
      IF ORD(func) # 0 THEN
        Out(GraphAddr, DataRotateReg);
        Out(GraphData, 18H);
      END;
    END SetFunc;

  PROCEDURE SetFuncWMode2(func: SYSTEM.BYTE);
    BEGIN
      IF ORD(func) # 0 THEN
        Out(GraphAddr, DataRotateReg);
        Out(GraphData, 18H);
      END;
      Out(GraphAddr, ModeReg);
      Out(GraphData, 2);
    END SetFuncWMode2;

  PROCEDURE SetFuncWMode2BitMask(func, m: SYSTEM.BYTE);
    BEGIN
      IF ORD(func) # 0 THEN
        Out(GraphAddr, DataRotateReg);
        Out(GraphData, 18H);
      END;
      Out(GraphAddr, ModeReg);
      Out(GraphData, 2);
      Out(GraphAddr, BitMaskReg);
      Out(GraphData, m);
    END SetFuncWMode2BitMask;

  PROCEDURE ResetFuncModeBitMask();
    BEGIN
      Out(GraphAddr, DataRotateReg);
      Out(GraphData, 0);
      Out(GraphAddr, ModeReg);
      Out(GraphData, 0);
      Out(GraphAddr, BitMaskReg);
      Out(GraphData, 0FFH);
    END ResetFuncModeBitMask;

  PROCEDURE SetFuncMapMask(func, m: SYSTEM.BYTE);
    BEGIN
      IF ORD(func) # 0 THEN
        Out(GraphAddr, DataRotateReg);
        Out(GraphData, 18H);
      END;
      Out(SeqAddr, MapMaskReg);
      Out(SeqData, m);
    END SetFuncMapMask;

  PROCEDURE ResetFuncMapMask();
    BEGIN
      Out(GraphAddr, DataRotateReg);
      Out(GraphData, 0);
      Out(SeqAddr, MapMaskReg);
      Out(SeqData, 0FH);
    END ResetFuncMapMask;

  PROCEDURE ResetFuncBitMapMask();
    BEGIN
      Out(GraphAddr, DataRotateReg);
      Out(GraphData, 0);
      Out(GraphAddr, BitMaskReg);
      Out(GraphData, 0FFH);
      Out(SeqAddr, MapMaskReg);
      Out(SeqData, 0FH);
    END ResetFuncBitMapMask;

  (* ================================ VGA-dependent control functions *)

  CONST
    NmbOfShades = 64;
    NmbOfColors = 256;
    NmbOfPureColors = 3;
    ColorTableLen = NmbOfColors * NmbOfPureColors;

  TYPE
    ColorTable = ARRAY [0 .. ColorTableLen - 1] OF SYSTEM.BYTE;
    ColorTablePtr = POINTER TO ColorTable;

  PROCEDURE VGAEnableVideo();
    BEGIN
      Out(SeqAddr, ClockingModeReg);
      Out(SeqData, SYSTEM.BYTE(BS(In(SeqData)) - BS{ScreenOffBit}));
    END VGAEnableVideo;

  PROCEDURE VGADisableVideo();
    BEGIN
      Out(SeqAddr, ClockingModeReg);
      Out(SeqData, SYSTEM.BYTE(BS(In(SeqData)) + BS{ScreenOffBit}));
    END VGADisableVideo;

  PROCEDURE VGASetColorReg(c, r, g, b: SYSTEM.BYTE);
    VAR
      f: CARDINAL;
    BEGIN
      WaitForRetrace();
      Out(DacAddrOut, c);
      Out(DacData, r);
      Out(DacData, g);
      Out(DacData, b);
    END VGASetColorReg;

  PROCEDURE VGASetAllColorRegs(a: SYSTEM.ADDRESS);
    VAR
      f, j: CARDINAL;
      ptr: ColorTablePtr;
    BEGIN
      ptr := a;
      WaitForRetrace();
      Out(DacAddrOut, 0);
      FOR j := 0 TO ColorTableLen - 1 DO
        Out(DacData, ptr^[j]);
      END;
    END VGASetAllColorRegs;

  PROCEDURE VGAGetAllColorRegs(a: SYSTEM.ADDRESS);
    VAR
      f, j: CARDINAL;
      ptr: ColorTablePtr;
    BEGIN
      ptr := a;
      WaitForRetrace();
      Out(DacAddrIn, 0);
      FOR j := 0 TO ColorTableLen - 1 DO
        ptr^[j] := SYSTEM.BYTE(ORD(In(DacData)) MOD NmbOfShades);
      END;
    END VGAGetAllColorRegs;

  PROCEDURE VGASetWMode1();
    BEGIN
      Out(GraphAddr, ModeReg);
      Out(GraphData, 41H);
    END VGASetWMode1;

  PROCEDURE VGAResetMode();
    BEGIN
      Out(GraphAddr, ModeReg);
      Out(GraphData, 40H);
    END VGAResetMode;

  PROCEDURE VGASetFuncWMode3(func, delta, c: SYSTEM.BYTE);
    BEGIN
      IF ORD(func) # 0 THEN delta := SYSTEM.BYTE(ORD(delta) + 18H) END;
      Out(GraphAddr, DataRotateReg);
      Out(GraphData, delta);
      Out(GraphAddr, ModeReg);
      Out(GraphData, 3);
      Out(GraphAddr, SetResetReg);
      Out(GraphData, c);
    END VGASetFuncWMode3;

  PROCEDURE VGASetPaletteSize16();
    BEGIN
      WaitForRetrace();
      Out(AttrAddr, ModeControlReg);
      Out(AttrDataOut, 81H);
      Out(AttrAddr, 20H);
    END VGASetPaletteSize16;

  PROCEDURE VGAResetPaletteSize();
    BEGIN
      WaitForRetrace();
      Out(AttrAddr, ModeControlReg);
      Out(AttrDataOut, 1);
      Out(AttrAddr, 20H);
    END VGAResetPaletteSize;

  PROCEDURE VGASetPalette(pal: SYSTEM.BYTE);
    BEGIN
      WaitForRetrace();
      Out(AttrAddr, ColorSelectReg);
      Out(AttrDataOut, pal);
      Out(AttrAddr, 20H);
    END VGASetPalette;

  PROCEDURE VGAPalette(): SYSTEM.BYTE;
    VAR
      pal: SYSTEM.BYTE;
    BEGIN
      WaitForRetrace();
      Out(AttrAddr, ColorSelectReg);
      pal := In(AttrDataIn);
      Out(AttrAddr, 20H);
      RETURN pal;
    END VGAPalette;

  (* ============================ Nonstandard and extended functions *)

  CONST
    CRTRegBlockLen = 24;

  TYPE
    CRTRegBlock = ARRAY [0 .. CRTRegBlockLen - 1] OF SYSTEM.BYTE;

  PROCEDURE InitCRT(misc: SYSTEM.BYTE; crt: CRTRegBlock);
    VAR
      in: CARDINAL;
    BEGIN
      (* ----------------------- reset the sequencer *)
      Out(SeqAddr, ResetReg);
      Out(SeqData, 1);
      (* -------------------------------------- set the misc register *)
      Out(MiscOut, misc);
      (* -------------------------------------------- set the CRT controller *)
      Out(CRTAddr, VerticalRetraceEndReg);
      Out(CRTData, SYSTEM.BYTE(BS(In(CRTData)) - BS{ProtectBit}));
      FOR in := 0 TO CRTRegBlockLen - 1 DO
        Out(CRTAddr, SYSTEM.BYTE(in));
        Out(CRTData, crt[in]);
      END;
      (* ----------------------------------- set the sequencer *)
      Out(SeqAddr, ResetReg);
      Out(SeqData, 3);
    END InitCRT;

  (* =============================================== Modified VGA *)

  (*%T CMPTS*)

  CONST
    CRTRegBlockM2 = CRTRegBlock
      (
        6DH,  5DH,  5EH,  8FH,  62H,  80H, 0BFH,  1FH,
         0H,  40H,   0H,   0H,   0H,   0H,   0H,   0H,
        83H,  85H,  5BH,  2FH,   0H,  63H, 0BAH, 0E3H
      );

  (*%E*)

  (*%T CMPSB*)

  CONST
    CRTRegBlockM2:CRTRegBlock =
      [
        6DH,  5DH,  5EH,  8FH,  62H,  80H, 0BFH,  1FH,
         0H,  40H,   0H,   0H,   0H,   0H,   0H,   0H,
        83H,  85H,  5BH,  2FH,   0H,  63H, 0BAH, 0E3H
      ];

  (*%E*)

  PROCEDURE ModifiedInit(m: CARDINAL);
    BEGIN
      IF (1 <= m) AND (m <= 2) THEN
        (* ------------------------------- force planes *)
        Out(SeqAddr, MemoryModeReg);
        Out(SeqData, 6);
        (* --------------------------------- set the CRT controller *)
        CASE m OF
        | 1:
          Out(CRTAddr, MaxScanLineReg);
          Out(CRTData, 40H);
          Out(CRTAddr, UnderlineLocReg);
          Out(CRTData, 0);
          Out(CRTAddr, CRTModeControlReg);
          Out(CRTData, 0E3H);
        | 2:
          InitCRT(0A7H, CRTRegBlockM2);
        END;
      END;
    END ModifiedInit;

  END ScrLow.
