IMPLEMENTATION MODULE SrvLex;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Initial version                         August 88
    A.Iakovlev  Ver  2.0  Full redesign                           July 89
    A.Iakovlev  Ver  2.1  Use M2VERS utility                      August 89
    A.Iakovlev  Ver  3.0  Renaming                                January 90
    A.Iakovlev  Ver  3.1  JPI-only version                        January 90
    A.Iakovlev  Ver  4.0  New processing of line-breaks           April 90
    A.Iakovlev  Ver  5.0  No ASCII                                July 90
    A.Iakovlev  Ver  5.1  No set difference (compiler error!)     July 90
    A.Iakovlev  Ver  6.0  Simplifications                         March 91
    A.Iakovlev  Ver  7.0  Pragma data(var_enum_size)              August 92
    A.Iakovlev  Ver  8.0  Renaming                                June 93
    A.Iakovlev  Ver  8.1  Import from SYSTEM                      January 94
    A.Iakovlev  Ver  9.0  Own type CharSet                        January 94
    A.Iakovlev  Ver 10.0  Stony Brook support                     February 94
    A.Iakovlev  Ver 11.0  Const MaxNmbOfDigits, no Str.StrToCard  February 94
    A.Iakovlev  Ver 11.1  No use Lib.ScanR, Lib.ScanNeR           February 94
    A.Iakovlev  Ver 12.0  Procedure PosChar                       February 94
    A.Iakovlev  Ver 13.0  No structured constants                 March 94
    *************************************************************************
    TopSpeed compiler dependencies:
    ------ pragmas (*#...*)
    ------ predefined type LONGCARD
    Stony Brook compiler dependencies:
    ------ predefined type LONGCARD
  *)

  IMPORT SYSTEM, SrvLow;

  CONST
    Space = " ";
    EndOfString = 0C;

  TYPE
    NumberDescriptor =
      RECORD
        digits: CharSet;
        base: CARDINAL;
        rightCh: CHAR;
      END;

  VAR NDA: ARRAY Base OF NumberDescriptor;

  PROCEDURE PosLineBreak(VAR str: ARRAY OF CHAR; inx: CARDINAL): CARDINAL;
    BEGIN
      INC(inx, SrvLow.ScanEqInc(SYSTEM.ADR(str[inx]), HIGH(str) - inx + 1, 15C));
      IF (inx < HIGH(str)) AND (str[inx + 1] = 12C) THEN
        RETURN inx;
      ELSIF inx = HIGH(str) THEN
        RETURN inx + 1;
      ELSE
        RETURN inx;
      END;
    END PosLineBreak;

  PROCEDURE PosChar(VAR str: ARRAY OF CHAR; inx: CARDINAL; ch: CHAR): CARDINAL;
    BEGIN
      INC(inx, SrvLow.ScanEqInc(SYSTEM.ADR(str[inx]), HIGH(str) - inx + 1, ch));
      IF inx = HIGH(str) THEN INC(inx) END;
      RETURN inx;
    END PosChar;

  PROCEDURE Skip(VAR str: ARRAY OF CHAR; inx: CARDINAL): CARDINAL;
    BEGIN
      RETURN inx + SrvLow.ScanNeInc(SYSTEM.ADR(str[inx]), HIGH(str) - inx + 1, Space);
    END Skip;

  PROCEDURE Extract(VAR str: ARRAY OF CHAR; inx: CARDINAL;
            allowed: CharSet; VAR result: ARRAY OF CHAR): CARDINAL;
    VAR
      k: CARDINAL;
    BEGIN
      k := 0;
      WHILE (inx <= HIGH(str)) AND (str[inx] IN allowed) DO
        IF k <= HIGH(result) THEN
          result[k] := str[inx];
          INC(k);
        END;
        INC(inx);
      END;
      IF k <= HIGH(result) THEN
        result[k] := EndOfString;
      END;
      RETURN inx;
    END Extract;

  PROCEDURE ExtractTypedNumber(VAR str: ARRAY OF CHAR; inx: CARDINAL;
            numType: Base; VAR x: CARDINAL): CARDINAL;
    VAR
      res: LONGCARD;
      inxSaved: CARDINAL;
      t: NumberDescriptor;
    BEGIN
      res := 0;
      inxSaved := inx;
      t := NDA[numType];
      WHILE (inx <= HIGH(str)) AND (str[inx] IN t.digits) DO
        IF inx - inxSaved < MaxNmbOfDigits THEN
          res := res * VAL(LONGCARD, t.base) + VAL(LONGCARD, ORD(str[inx]) - ORD("0"));
        END;
        INC(inx);
      END;
      IF (numType = dec) OR (str[inx] = t.rightCh) THEN
        x := VAL(CARDINAL, res);
        IF numType = dec THEN
          RETURN inx;
        ELSE
          RETURN inx + 1;
        END;
      ELSE
        RETURN inxSaved;
      END;
    END ExtractTypedNumber;

  PROCEDURE ExtractNumber(VAR str: ARRAY OF CHAR; inx: CARDINAL;
            VAR x: CARDINAL): CARDINAL;
    VAR
      i: CARDINAL;
      type: Base;
    BEGIN
      FOR type := MIN(Base) TO MAX(Base) DO
        i := ExtractTypedNumber(str, inx, type, x);
        IF i > inx THEN
          RETURN i;
        END;
      END;
      RETURN inx;
    END ExtractNumber;

  BEGIN
    WITH NDA[dec] DO
      digits  := CharSet{"0" .. "9"};
      base    := 10;
      rightCh := " ";
    END;
    WITH NDA[hex] DO
      digits  := CharSet{"0" .. "9", "A" .. "F"};
      base    := 16;
      rightCh := "H";
    END;
    WITH NDA[oct] DO
      digits  := CharSet{"0" .. "7"};
      base    := 8;
      rightCh := "B";
    END;
  END SrvLex.
